from pathlib import Path
from typing import Tuple
import json

from amino import Maybe, Empty, Just, List, Map, may, Boolean, flat_may, __, F
from amino.lazy import lazy

from fn import _

from ribosome.nvim import NvimFacade, HasNvim
from ribosome.record import field, list_field, Record
from ribosome.process import JobClient

from proteome.logging import Logging


def mkpath(path: str):
    return Path(path).expanduser()  # type: ignore


def format_path(path: Path):
    h = str(Path.home())  # type: ignore
    return str(path).replace(h, '~')


# TODO subprojects, e.g. sbt projects
class Project(Record):
    name = field(str)
    root = field(Path)
    tpe = field(Maybe, initial=Empty())
    types = list_field()
    langs = list_field()
    history = field(bool, initial=False)

    @staticmethod
    def of(name, root, tpe=Empty(), **kw):
        return Project(name=name, root=root, tpe=tpe, **kw)

    @property
    def ident(self):
        return self.tpe.map(_ + '/').get_or_else('') + self.name

    @property
    def info(self) -> str:
        return '{}: {}'.format(self.ident, format_path(self.root))

    def __str__(self):
        return self.info

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self.ident))

    def __eq__(self, other):
        return (
            isinstance(other, Project) and
            self.name == other.name and
            self.root == other.root and
            self.tpe == other.tpe
        )

    def __hash__(self):
        return hash((self.name, self.root, self.tpe))

    @property
    def ctags_langs(self):
        return (self.langs + self.tpe.to_list).distinct

    @property
    def want_ctags(self):
        return not self.ctags_langs.is_empty

    @property
    def tag_file(self):
        return self.root / '.tags'

    def remove_tag_file(self):
        if self.tag_file.exists():
            self.tag_file.unlink()

    @property
    def fqn(self):
        t = self.tpe.get_or_else('notype')
        return '{}__{}'.format(t, self.name)

    @property
    def json(self):
        return dict(
            name=self.name,
            root=str(self.root),
            tpe=self.tpe | None,
            types=self.types,
            langs=self.langs,
        )

    @property
    def all_types(self):
        return self.tpe.to_list + self.types

    @property
    def has_type(self):
        return self.tpe.is_just

    @lazy
    def job_client(self):
        return JobClient(cwd=self.root, name=self.ident)

    def match_ident(self, ident):
        return ident in (self.ident, self.name)


class Projects(object):

    def __init__(self, projects: List[Project]=List()) -> None:
        self.projects = projects

    def __add__(self, pro: Project) -> 'Projects':
        return Projects(self.projects + [pro])

    def __sub__(self, pro: Project) -> 'Projects':
        return Projects(self.projects.without(pro))

    def __pow__(self, pros: List[Project]) -> 'Projects':
        return Projects(self.projects + pros)

    def show(self, names: List[str]=List()):
        if names.is_empty:
            pros = self.projects
        else:
            pros = names.flat_map(self.project)
        return pros.map(_.info)

    def project(self, ident: str) -> Maybe[Project]:
        @flat_may
        def try_split():
            if '/' in ident:
                tpe, name = ident.split('/', 1)
                return self.projects.find(
                    lambda a: a.name == name and a.tpe == Just(tpe))
        return self.projects.find(_.name == ident)\
            .or_else(try_split)

    def ctags(self, names: List[str]):
        matching = names.flat_map(self.project)
        return matching

    def __str__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ','.join(map(repr, self.projects))
        )

    def __getitem__(self, index):
        return self.projects.lift(index)

    def __len__(self):
        return len(self.projects)

    def __contains__(self, item):
        return (
            (isinstance(item, Project) and item in self.projects) or
            (isinstance(item, str) and self.project(item).is_just)
        )

    @property
    def json(self):
        return self.projects.map(_.json)

    def index_of(self, project):
        return self.projects.index_of(project)

    def index_of_ident(self, ident):
        return self.projects.index_where(_.ident == ident)\
            .or_else(self.projects.index_where(_.name == ident))

    @property
    def idents(self):
        return self.projects.map(_.ident)


def sub_path(base: Path, path: Path):
    check = lambda: path.relative_to(str(base))
    return Maybe.from_call(check, exc=ValueError)\
        .map(_.parts)\
        .map(List.wrap)


def sub_paths(dirs: List[Path], path: Path):
    return dirs\
        .flat_map(lambda a: sub_path(a, path))


class Resolver(object):

    def __init__(self, bases: List[Path], types: Map[Path, List[str]]) -> None:
        self.bases = bases
        self.types = types

    def type_name(self, tpe: str, name: str) -> Maybe[Path]:
        return self.bases\
            .map(_ / tpe / name)\
            .find(lambda a: a.is_dir())\
            .or_else(lambda: self.specific(tpe, name))

    def specific(self, tpe: str, name: str) -> Maybe[Path]:
        return self.types\
            .valfilter(_.call('contains', tpe))\
            .k\
            .map(_ / name)\
            .find(lambda a: a.is_dir())

    def dir(self, path: Path) -> Maybe[Tuple[str, str]]:
        return self.dir_in_bases(path)\
            .or_else(lambda: self.dir_in_types(path))

    def dir_in_bases(self, path: Path) -> Maybe[Tuple[str, str]]:
        return sub_paths(self.bases, path)\
            .filter(lambda a: len(a) >= 2)\
            .map(lambda a: (a[0], a[-1]))\
            .head

    def dir_in_types(self, path: Path) -> Maybe[Tuple[str, str]]:
        trans = lambda a, b: b.head.zip(sub_path(a, path).flat_map(_.last))
        return self.types\
            .flat_map(trans)\
            .head


def content(path: Path):
    return List.wrap(path.iterdir()) if path.is_dir() else List()


def subdirs(path: Path, n: int):
    sub = content(path).filter(__.is_dir())
    return sub if n <= 1 else sub // F(subdirs, n=n - 1)


def extract_ident(path: Path):
    return (
        Just(path.parts)
        .filter(lambda a: len(a) >= 2)
        .map(lambda a: '/'.join(a[-2:]))
    )


class ProjectLoader(Logging):

    def __init__(self, config_path: Path, resolver: Resolver) -> None:
        self.resolver = resolver
        self.config_path = config_path
        self.config = self._load_config()

    def _load_config(self) -> List[Map]:
        def parse(path: Path):
            with path.open() as f:
                try:
                    return List.wrap(map(Map, json.loads(f.read())))
                except Exception as e:
                    self.log.error('parse error in {}: {}'.format(path, e))
                    return List()
        if (self.config_path.is_dir()):
            return List.wrap(self.config_path.glob('*.json')) \
                .flat_map(parse)
        elif self.config_path.is_file():
            return parse(self.config_path)
        else:
            return List()

    def resolve(self, tpe: str, name: str):
        return self.resolver.type_name(tpe, name)\
            .map(lambda a: Project.of(name, a, Just(tpe)))

    def resolve_ident(self, ident: str, params: Map=Map(),
                      main: Maybe[str]=Empty()):
        if '/' in ident:
            return self.resolve(*ident.split('/', 1))
        else:
            return main.flat_map(lambda a: self.resolve(a, ident))

    def json_by_name(self, name: str):
        return self.config\
            .find(lambda a: a.get('name').contains(name))

    def json_by_type_name(self, tpe, name):
        def matcher(record: Map) -> Boolean:
            return (record.get('name').contains(name) and
                    record.get('type').contains(tpe))
        return self.config\
            .find(matcher)

    def json_by_ident(self, ident: str):
        @flat_may
        def try_split():
            if '/' in ident:
                return self.json_by_type_name(*ident.split('/', 1))
        return self.json_by_name(ident)\
            .or_else(try_split)

    def json_by_root(self, root: Path):
        return self.config\
            .find(lambda a: a.get('root').map(mkpath).contains(root))

    def by_ident(self, name: str):
        return self.json_by_ident(name)\
            .flat_map(self.from_json)

    def from_json(self, json: Map) -> Maybe[Project]:
        ''' Try to instantiate a Project from the given json object.
        Convert the **type** key to **tpe** and its value to
        Maybe.
        Make sure **root** is a directory, fall back to resolution
        by **tpe/name**.
        Reinsert the root dir into the json dict, filter out all keys
        that aren't contained in Project's fields.
        Try to instantiate.
        '''
        root = json.get('root')\
            .map(mkpath)\
            .or_else(
                json.get_all('type', 'name')
                .flat_smap(self.resolver.type_name))
        valid_fields = root\
            .map(lambda a: json ** Map(root=a, tpe=json.get('type')))\
            .map(lambda a: a.at(*Project._pclass_fields))
        return Maybe.from_call(lambda: valid_fields.ssmap(Project)) | Empty()

    # TODO log error if not a dir
    # use Either
    @may
    def create(self, name: str, root: Path, **kw):
        if root.expanduser().is_dir():  # type: ignore
            return Project.of(name, root, **kw)

    def from_params(self, ident: str, root: Path, params: Map):
        parts = List(*reversed(ident.split('/', 1)))
        name = parts[0]
        tpe = parts.lift(1).or_else(params.get('type'))
        kw = params.at('types', 'langs', 'history')
        return self.create(name, root, tpe=tpe, **kw)

    @property
    def _all_long_ident(self):
        def typed_ident(base, types):
            names = subdirs(base, n=1) / _.name
            return (types // (lambda t: names / F('{}/{}'.format, t)))
        bases = self.resolver.bases // F(subdirs, n=2) // extract_ident
        typed = self.resolver.types.to_list.flat_smap(typed_ident)
        return bases + typed

    def _short_ident(self, idents):
        return idents / __.split('/') / _[-1]

    def _main_ids(self, idents, main: Maybe[str]):
        return main / (
            lambda a: idents.filter(__.startswith(a + '/'))) | List()

    def all_ident(self, main: Maybe[str]):
        all = self._all_long_ident
        m_ids = self._main_ids(all, main)
        return all + self._short_ident(m_ids)

    def main_ident(self, main: Maybe[str]):
        m_ids = self._main_ids(self._all_long_ident, main)
        return self._short_ident(m_ids)


class ProjectAnalyzer(HasNvim, Logging):

    def __init__(self, vim: NvimFacade, loader: ProjectLoader) -> None:
        super(ProjectAnalyzer, self).__init__(vim)
        self.loader = loader

    def _default_detect_data(self, wd: Path):
        type_name = self.loader.resolver.dir(wd)
        return type_name\
            .flat_smap(self.loader.json_by_type_name)\
            .map(_ + ('root', str(wd)))\
            .or_else(
                type_name
                .smap(lambda t, n: Map(type=t, root=str(wd), name=n))
            )

    def _detect_data(self, wd: Path):
        return self.loader.json_by_root(wd)\
            .or_else(
                self.vim.pvar('project_detector')
                .flat_map(lambda a: self.vim.call(a, str(wd))))\
            .or_else(self._default_detect_data(wd))

    @property
    def main_dir(self) -> Maybe[Path]:
        return Maybe.from_call(Path.cwd, exc=IOError)

    @property
    def main_dir_or_home(self) -> Path:
        return self.main_dir | Path.home()  # type: ignore

    @property
    def _detect_main_data(self) -> Maybe[Map]:
        return self.main_dir // self._detect_data

    @property
    def _main_data(self) -> Maybe[Map]:
        return self.pflags.get('detect_main_project', True)\
            .maybe(self._detect_main_data) | Empty()

    @property
    def _auto_main(self):
        return self._main_data\
            .flat_map(self.loader.from_json)

    @property
    def _fallback_main(self):
        tpe = self.vim.pvar('main_project_type')
        return Project.of('main', self.main_dir_or_home, tpe)

    @property
    def main(self):
        return self.vim.pvar('main_project')\
            .flat_map(self.loader.by_ident)\
            .or_else(self._auto_main)\
            .get_or_else(self._fallback_main)

__all__ = ('Projects', 'Project', 'ProjectLoader', 'Resolver')



from proteome.plugins.unite.main import Plugin
from proteome.plugins.unite.data import (UniteSelectAdd, UniteSelectAddAll,
                                         UniteProjects, Id)

__all__ = ('Plugin', 'UniteSelectAdd', 'UniteSelectAddAll', 'UniteProjects',
           'Id')



from pathlib import Path

import neovim

import ribosome

from proteome.logging import Logging


class NvimFacade(Logging, ribosome.nvim.NvimFacade):

    def __init__(self, nvim: neovim.Nvim) -> None:
        ribosome.nvim.NvimFacade.__init__(self, nvim, 'proteome')

    def cd(self, d):
        self.vim.chdir(str(d))

    def switch_root(self, path):
        p = str(path)
        self.cd(p)
        self.pautocmd('SwitchedRoot')
        self.set_pvar('root_dir', p)


__all__ = ['NvimFacade']



from ribosome.machine import may_handle

from amino import List, Map

from proteome.state import ProteomeComponent, ProteomeTransitions
from proteome.plugins.unite.data import (UniteSelectAdd, UniteSelectAddAll,
                                         UniteProjects, UniteSource, UniteKind,
                                         Id, UniteMessage)


class Plugin(ProteomeComponent):

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._unite_ready = False

    def prepare(self, msg):
        if (not self._unite_ready and isinstance(msg, UniteMessage)):
            self._setup_unite()

    def _setup_unite(self):
        addable = UniteSource(Id.addable, Id.addable_candidates, Id.addable)
        all_addable = UniteSource(Id.all_addable, Id.all_addable_candidates,
                                  Id.addable)
        add_action = Map(name='add', handler=Id.add_project,
                         desc='add project')
        add_pro = UniteKind(Id.addable, List(add_action))
        projects = UniteSource(Id.projects, Id.projects_candidates, Id.project)
        delete_action = Map(name='delete', handler=Id.delete_project,
                            desc='delete project')
        activate_action = Map(name='activate', handler=Id.activate_project,
                              desc='activate project', is_selectable=0)
        project = UniteKind(Id.project, List(activate_action, delete_action))
        addable.define(self.vim)
        all_addable.define(self.vim)
        add_pro.define(self.vim)
        projects.define(self.vim)
        project.define(self.vim)
        self._unite_ready = True

    class Transitions(ProteomeTransitions):

        def unite_cmd(self, cmd):
            args = ' '.join(self.msg.unite_args)
            self.vim.cmd('Unite {} {}'.format(cmd, args))

        @may_handle(UniteSelectAdd)
        def select_add(self):
            self.unite_cmd(Id.addable)

        @may_handle(UniteSelectAddAll)
        def select_add_all(self):
            self.unite_cmd(Id.all_addable)

        @may_handle(UniteProjects)
        def projects(self):
            self.unite_cmd(Id.projects)

__all__ = ('Plugin')



from fn import _

from ribosome.machine import may_handle, message
from ribosome.nvim import Buffer

from amino import List, Empty
from amino.lazy import lazy

from proteome.state import ProteomeComponent, ProteomeTransitions
from proteome.ctags import Ctags
from proteome.plugins.core import Save, Added, BufEnter, StageIV

Gen = message('Gen', 'project')
GenAll = message('GenAll')
Kill = message('Kill')
CurrentBuffer = message('CurrentBuffer')


class Plugin(ProteomeComponent):

    @lazy
    def ctags(self):
        return Ctags(self.vim)

    class Transitions(ProteomeTransitions):

        @property
        def ctags(self) -> Ctags:
            return self.machine.ctags  # type: ignore

        @property
        def _tags_file_name(self):
            return '.tags'

        @may_handle(StageIV)
        def stage_4(self):
            return GenAll(), CurrentBuffer()

        @may_handle(Save)
        def save(self):
            return GenAll()

        @may_handle(GenAll)
        def gen_all(self):
            if self.ctags.ready:
                return self.data.projects.projects\
                    .filter(_.want_ctags)\
                    .map(Gen)

        # TODO kill dangling procs
        @may_handle(Kill)
        def kill(self):
            pass

        @may_handle(BufEnter)
        def buf_enter(self):
            self.set_buffer_tags(List(self.msg.buffer))

        @may_handle(Added)
        def added(self: Added):
            return CurrentBuffer()

        @may_handle(CurrentBuffer)
        def buffer(self):
            self.set_buffer_tags(List(self.vim.buffer))

        def set_buffer_tags(self, bufs: List[Buffer]):
            files = self.data.all_projects.map(_.root / self._tags_file_name)
            bufs.foreach(lambda a: a.amend_optionl('tags', files))

        @may_handle(Gen)
        async def gen(self):
            await self.ctags.gen(self.msg.project)
            return Empty()

__all__ = ('GenAll', 'Plugin')



from proteome.git.process import Git
from proteome.git.repo import (Repo, RepoT, DulwichRepo, RepoAdapter,
                               CommitInfo, RepoState, ProjectRepoAdapter)

__all__ = ('Git', 'Repo', 'RepoT', 'DulwichRepo', 'RepoAdapter', 'CommitInfo',
           'RepoState', 'ProjectRepoAdapter')



from pathlib import Path

import neovim

from amino import List, Map, __, F, _

from ribosome import command, NvimStatePlugin, msg_command, json_msg_command

from proteome.plugins.core import (
    AddByParams, Show, Create, SetProject, Next, Prev, StageI,
    Save, RemoveByIdent, BufEnter, StageII, StageIII, StageIV, CloneRepo
)
from proteome.plugins.history.messages import (HistoryPrev, HistoryNext,
                                               HistoryStatus, HistoryLog,
                                               HistoryBrowse,
                                               HistoryBrowseInput,
                                               HistorySwitch, HistoryPick,
                                               HistoryRevert,
                                               HistoryFileBrowse)
from proteome.main import Proteome
from proteome.nvim import NvimFacade
from proteome.logging import Logging
from proteome.plugins.unite import (UniteSelectAdd, UniteSelectAddAll,
                                    UniteProjects)
from proteome.plugins.unite import Id as Unite


def unite_candidates(name):
    handler = getattr(Unite, '{}_candidates'.format(name))
    def uc_wrap(f):
        @neovim.function(handler, sync=True)
        def f_wrap(*a, **kw):
            return f(*a, **kw) / (lambda a: {'word': a})
        return f_wrap
    return uc_wrap


def _unite_word(args):
    first = List.wrap(args).lift(0)
    candidates = (
        first.to_list.join
        if first.exists(lambda a: isinstance(a, list))
        else first
    )
    return candidates / Map // __.get('word')


def unite_action(name):
    handler = getattr(Unite, name)
    def uc_wrap(f):
        @neovim.function(handler)
        def f_wrap(self, args):
            _unite_word(args) / F(f, self) % self.pro.send
        return f_wrap
    return uc_wrap


class ProteomeNvimPlugin(NvimStatePlugin, Logging):

    def __init__(self, vim: neovim.Nvim) -> None:
        super().__init__(NvimFacade(vim))
        self.pro = None
        self._post_initialized = False

    @property
    def state(self):
        return self.pro

    @command()
    def proteome_reload(self):
        self.proteome_quit()
        self.proteome_start()
        self._post_startup()

    @command()
    def proteome_quit(self):
        if self.pro is not None:
            self.vim.clean()
            self.pro.stop()
            self.pro = None

    @command(sync=True)
    def proteome_start(self):
        config_path = self.vim.ppath('config_path')\
            .get_or_else(Path('/dev/null'))
        bases = self.vim.ppathl('base_dirs')\
            .get_or_else(List())\
            .map(Path)
        type_bases = self.vim.pd('type_base_dirs')\
            .get_or_else(Map())\
            .keymap(lambda a: Path(a).expanduser())\
            .valmap(List.wrap)
        plugins = self.vim.pl('plugins') | List()
        self.pro = Proteome(self.vim.proxy, Path(config_path), plugins, bases,
                            type_bases)
        self.pro.start()
        self.pro.wait_for_running()
        self.pro.send(StageI())

    @neovim.autocmd('VimEnter')
    def vim_enter(self):
        if not self._post_initialized:
            self.proteome_post_startup()

    @command()
    def proteome_post_startup(self):
        self._post_initialized = True
        if self.pro is not None:
            self.pro.send(StageII().at(1))
            self.pro.send(StageIII().at(1))
            self.pro.send(StageIV().at(1))
        else:
            self.log.error('proteome startup failed')

    @command()
    def pro_plug(self, plug_name, cmd_name, *args):
        self.pro.plug_command(plug_name, cmd_name, args)

    @msg_command(Create)
    def pro_create(self):
        pass

    addable = dict(complete='customlist,ProCompleteAddableProjects')
    projects = dict(complete='customlist,ProCompleteProjects')

    @json_msg_command(AddByParams, **addable)
    def pro_add(self):
        pass

    @neovim.function('ProCompleteProjects', sync=True)
    def pro_complete_projects(self, args):
        lead, line, pos = args
        return self.pro.data.projects.idents.filter(__.startswith(lead))

    @neovim.function('ProCompleteAddableProjects', sync=True)
    def pro_complete_addable_projects(self, args):
        lead, line, pos = args
        return self.pro.data.addable.filter(__.startswith(lead))

    @msg_command(RemoveByIdent, **projects)
    def pro_remove(self):
        pass

    @msg_command(Show)
    def pro_show(self):
        pass

    @msg_command(SetProject, **projects)
    def pro_to(self):
        pass

    @msg_command(Next)
    def pro_next(self):
        pass

    @msg_command(Prev)
    def pro_prev(self):
        pass

    @msg_command(Save)
    def pro_save(self):
        pass

    # TODO start terminal at root dir
    # @msg_command(Term)
    # def pro_term(self):
        # pass

    @neovim.autocmd('BufEnter')
    def buf_enter(self):
        if self._post_initialized:
            self.pro.send(BufEnter(self.vim.buffer.proxy))

    @json_msg_command(CloneRepo)
    def pro_clone(self):
        pass

    @msg_command(HistoryPrev)
    def pro_history_prev(self):
        pass

    @msg_command(HistoryNext)
    def pro_history_next(self):
        pass

    @msg_command(HistoryStatus)
    def pro_history_status(self):
        pass

    @msg_command(HistoryLog)
    def pro_history_log(self):
        pass

    @msg_command(HistoryBrowse)
    def pro_history_browse(self):
        pass

    @msg_command(HistoryFileBrowse)
    def pro_history_file_browse(self):
        pass

    @msg_command(HistoryBrowseInput)
    def pro_history_browse_input(self):
        pass

    @msg_command(HistorySwitch)
    def pro_history_switch(self):
        pass

    @msg_command(HistoryPick)
    def pro_history_pick(self):
        pass

    @msg_command(HistoryRevert)
    def pro_history_revert(self):
        pass

    @msg_command(UniteSelectAdd)
    def pro_select_add(self):
        pass

    @msg_command(UniteSelectAddAll)
    def pro_select_add_all(self):
        pass

    @msg_command(UniteProjects)
    def projects(self):
        pass

    @unite_candidates('addable')
    def pro_unite_addable(self, args):
        return self.pro.data.main_addable

    @unite_candidates('all_addable')
    def pro_unite_all_addable(self, args):
        return self.pro.data.addable

    @unite_candidates('projects')
    def pro_unite_projects(self, args):
        return self.pro.data.all_projects / _.name

    @unite_action('add_project')
    def pro_unite_add_project(self, ident):
        return AddByParams(ident, params=Map())

    @unite_action('activate_project')
    def pro_unite_activate_project(self, ident):
        return SetProject(ident)

    @unite_action('delete_project')
    def pro_unite_delete_project(self, ident):
        return RemoveByIdent(ident)

__all__ = ('ProteomeNvimPlugin',)
